// Package openapi2 parses and writes OpenAPIv2 specification documents.
//
// Does not cover all elements of OpenAPIv2.
// When OpenAPI version 3 is backwards-compatible with version 2, version 3 elements have been used.
//
// See https://github.com/OAI/OpenAPI-Specification/blob/master/versions/2.0.md
package openapi2
